////////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2018 Josef Spidlen, Ph.D., FlowJo, LLC
//
// License
// The software is distributed under the terms of the 
// Artistic License 2.0
// http://www.r-project.org/Licenses/Artistic-2.0
// 
// Disclaimer
// This software and documentation come with no warranties of any kind.
// This software is provided "as is" and any express or implied 
// warranties, including, but not limited to, the implied warranties of
// merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the  copyright holder be liable for any direct, 
// indirect, incidental, special, exemplary, or consequential damages
// (including but not limited to, procurement of substitute goods or 
// services; loss of use, data or profits; or business interruption)
// however caused and on any theory of liability, whether in contract,
// strict liability, or tort arising in any way out of the use of this 
// software.    
//////////////////////////////////////////////////////////////////////////////

package com.flowjo.plugin.FitSNE;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import com.treestar.lib.file.FileUtil;
import com.flowjo.plugin.FitSNE.utils.FilenameUtils;
import com.treestar.flowjo.engine.EngineManager;
import com.treestar.flowjo.engine.utility.RFlowCalculator;

public class FItSNERFlowCalc extends RFlowCalculator {
	
	private final static String scriptTemplatePath = "r/RScript.FItSNE.Template.R";
    public static final String TRUE = "TRUE";
    public static final String FALSE = "FALSE";

	
	public File runAlg(File sampleFile, String sampleName, List<String> parameterNames, Map<String, String> options, String outputFolderPath, boolean useExistingFiles, String prefix)
    {
		File algScript = null;
		try
		{
			sampleName = sampleName.replaceAll(".ExtNode", "").replaceAll(".fcs", "").replaceAll(".LMD", "").trim();
			File outputFolder = new File(outputFolderPath);
			File executable = getExecutable(outputFolder);
			if (executable == null)
				return null;
			StringWriter scriptWriter = new StringWriter();
			algScript = createAlgScript(sampleFile, sampleName, parameterNames, options, outputFolder, executable, scriptWriter, prefix);
			if (algScript == null)
				return null;
			if (useExistingFiles && algScript.exists())
				return new File(algScript, FItSNE.alreadyExists);

			String scriptFileName = (new StringBuilder()).append("RScript.FItSNE.").append(System.currentTimeMillis()).append(".R").toString().replaceAll(" ", "_");
			fScriptFile = new File(outputFolderPath, scriptFileName);
			FileUtil.write(fScriptFile, scriptWriter.toString());
			executeRBatch(scriptFileName);
		}
		catch (Exception e)
		{
			e.printStackTrace();
		}
		return algScript;
    }

	// Extract the required binaries from the jar to the output folder; return the name of the executable
    protected File getExecutable(File outputFolder) throws Exception {
    	
        InputStream stream = null;
        OutputStream resStreamOut = null;
        File outFile = null;
        File retFile = null;
        
        String[] resources;
        if (EngineManager.isWindows()) 
        	resources = FItSNE.WinResources;
        else
        	resources = FItSNE.OSXResources;
        
        try {
        	
        	for (String resourceName : resources) {
        		outFile = new File(outputFolder, resourceName);
        		if (outFile.exists()) {
        			System.out.println(outFile.getAbsolutePath() + " already exists, will not be extracting again.");
        		}
        			
        		stream = FItSNERFlowCalc.class.getResourceAsStream(resourceName);
                if (stream == null) {
                    throw new Exception("Failed to extract resource \"" + resourceName + "\" from the jar file.");
                }
                
                int readBytes;
                byte[] buffer = new byte[4096];
                if (!outFile.getParentFile().exists())
                	outFile.getParentFile().mkdirs();
                resStreamOut = new FileOutputStream(outFile);
                while ((readBytes = stream.read(buffer)) > 0) {
                    resStreamOut.write(buffer, 0, readBytes);
                }
                if (resStreamOut != null) {
                    resStreamOut.close();
                    resStreamOut = null;
                }
                if (stream != null) {
                	stream.close();
                	stream = null;
                }
                    
                outFile.setExecutable(true);
                if (retFile == null)
                	retFile = outFile;

        	}
        } catch (Exception ex) {
            System.out.println("Exception when extracting resources from jar.");
            throw ex;
        } finally {
            if (resStreamOut != null) {
                resStreamOut.close();
            }
            if (stream != null) {
                stream.close();
            }
        }
        
        System.out.println("Executable binaries extracted to " + outFile);
        return retFile;
    }
	
	protected File createAlgScript(File sampleFile, String sampleName, List<String> parameterNames, Map<String, String> options, File outputFolder, File executable, StringWriter scriptWriter, String prefix)
    {
        InputStream scriptStream = FItSNERFlowCalc.class.getResourceAsStream(scriptTemplatePath);
        if(outputFolder == null) outputFolder = sampleFile.getParentFile();

        String dataFilePath = sampleFile.getAbsolutePath();
        String executablePath = executable.getAbsolutePath();

        String sAddCellIdToResults = FItSNE.isSeqGeq() ? TRUE : FALSE;
        String sNoMomentumDuringExag = options.get(FItSNE.noMomentumDuringExagOptionName);
        if (sNoMomentumDuringExag != null && !sNoMomentumDuringExag.isEmpty() && (FItSNE.True.equals(sNoMomentumDuringExag) || FItSNE.One.equals(sNoMomentumDuringExag)))
        	sNoMomentumDuringExag = TRUE;
        else
        	sNoMomentumDuringExag = FALSE;
        
        String sDimsOption = options.get(FItSNE.dimOptionName);
        String sAlgOption = options.get(FItSNE.algOptionName);
        String sKnnAlgOption = options.get(FItSNE.knnAlgOptionName);

        String sDerParName = options.get(FItSNE.derParNameOptionName);
        if (sDerParName.isEmpty() ||  FItSNE.derParNameDefault.equalsIgnoreCase(sDerParName)) 
        	sDerParName = "";	// Empty string will signal to R that it should come up with its own name.
        
        String sPerplexity = options.get(FItSNE.perplexityOptionName);
        try {
        	if (Double.isNaN(Double.parseDouble(sPerplexity)) || (Double.parseDouble(sPerplexity) < FItSNE.perplexityMin || Double.parseDouble(sPerplexity) > FItSNE.perplexityMax))
        		sPerplexity = Double.toString(FItSNE.perplexityDefault);
        } catch (Exception e) {
        	sPerplexity = Double.toString(FItSNE.perplexityDefault);
        }

        String sMaxIter = options.get(FItSNE.maxIterOptionName);
        try {
        	if ((Integer.parseInt(sMaxIter) < FItSNE.maxIterMin) || (Integer.parseInt(sMaxIter) > FItSNE.maxIterMax))
        		sMaxIter = Integer.toString(FItSNE.maxIterDefault);
        } catch (Exception e) {
        	sMaxIter = Integer.toString(FItSNE.maxIterDefault);
        }

        String sStopLyingIter = options.get(FItSNE.stopLyingIterOptionName);
        try {
        	if ((Integer.parseInt(sStopLyingIter) < FItSNE.stopLyingMin) || (Integer.parseInt(sStopLyingIter) > FItSNE.stopLyingMax))
        		sStopLyingIter = Integer.toString(FItSNE.stopLyingDefault);
        } catch (Exception e) {
        	sStopLyingIter = Integer.toString(FItSNE.stopLyingDefault);
        }
        
        String sExaggerationFactor = options.get(FItSNE.exaggerationFactorOptionName);
        try {
        	if (Double.isNaN(Double.parseDouble(sExaggerationFactor)) || (Double.parseDouble(sExaggerationFactor) < FItSNE.exaggerationFactorMin || Double.parseDouble(sExaggerationFactor) > FItSNE.exaggerationFactorMax))
        		sExaggerationFactor = Double.toString(FItSNE.exaggerationFactorDefault);
        } catch (Exception e) {
        	sExaggerationFactor = Double.toString(FItSNE.exaggerationFactorDefault);
        }

        String sTheta = options.get(FItSNE.thetaOptionName);
        try {
        	if (Double.isNaN(Double.parseDouble(sTheta)) || (Double.parseDouble(sTheta) < FItSNE.thetaMin || Double.parseDouble(sTheta) > FItSNE.thetaMax))
        		sTheta = Double.toString(FItSNE.thetaDefault);
        } catch (Exception e) {
        	sTheta = Double.toString(FItSNE.thetaDefault);
        }

        String sLateEgCoeff = options.get(FItSNE.lateEgCoeffOptionName);
        try {
        	if (Double.isNaN(Double.parseDouble(sLateEgCoeff)) || (Double.parseDouble(sLateEgCoeff) < FItSNE.lateEgCoeffMin || Double.parseDouble(sLateEgCoeff) > FItSNE.lateEgCoeffMax))
        		sLateEgCoeff = Double.toString(FItSNE.lateEgCoeffDefault);
        } catch (Exception e) {
        	sLateEgCoeff = Double.toString(FItSNE.lateEgCoeffDefault);
        }
        
        String sStartLateExag = options.get(FItSNE.startLateExagIterOptionName);
        try {
        	if ((Integer.parseInt(sStartLateExag) < FItSNE.startLateExagIterMin) || (Integer.parseInt(sStopLyingIter) > FItSNE.startLateExagIterMax))
        		sStartLateExag = Integer.toString(FItSNE.startLateExagIterDefault);
        } catch (Exception e) {
        	sStartLateExag = Integer.toString(FItSNE.startLateExagIterDefault);
        }

        String sNumTrees = options.get(FItSNE.numTreesOptionName);
        try {
        	if ((Integer.parseInt(sNumTrees) < FItSNE.numTreesMin) || (Integer.parseInt(sNumTrees) > FItSNE.numTreesMax))
        		sNumTrees = Integer.toString(FItSNE.numTreesDefault);
        } catch (Exception e) {
        	sNumTrees = Integer.toString(FItSNE.numTreesDefault);
        }
        
        String sNumTerms = options.get(FItSNE.numTermsOptionName);
        try {
        	if ((Integer.parseInt(sNumTerms) < FItSNE.numTermsMin) || (Integer.parseInt(sNumTerms) > FItSNE.numTermsMax))
        		sNumTerms = Integer.toString(FItSNE.numTermsDefault);
        } catch (Exception e) {
        	sNumTerms = Integer.toString(FItSNE.numTermsDefault);
        }
        
        String sSearchK = options.get(FItSNE.searchKOptionName);
        try {
        	if ((Integer.parseInt(sSearchK) < FItSNE.searchKMin) || (Integer.parseInt(sSearchK) > FItSNE.searchKMax))
        		sSearchK = Integer.toString(FItSNE.searchKDefault);
        } catch (Exception e) {
        	sSearchK = Integer.toString(FItSNE.searchKDefault);
        }
        
        String sK = options.get(FItSNE.kOptionName);
        try {
        	if ((Integer.parseInt(sK) < FItSNE.kMin) || (Integer.parseInt(sK) > FItSNE.kMax))
        		sK = Integer.toString(FItSNE.kDefault);
        } catch (Exception e) {
        	sK = Integer.toString(FItSNE.kDefault);
        }
        
        String sSigma = options.get(FItSNE.sigmaOptionName);
        try {
        	if (Double.isNaN(Double.parseDouble(sSigma)) || (Double.parseDouble(sSigma) < FItSNE.sigmaMin || Double.parseDouble(sSigma) > FItSNE.sigmaMax))
        		sSigma = Double.toString(FItSNE.sigmaDefault);
        } catch (Exception e) {
        	sSigma = Double.toString(FItSNE.sigmaDefault);
        }

        String sMinNumIntervals = options.get(FItSNE.minNumIntervalsOptionName);
        try {
        	if ((Integer.parseInt(sMinNumIntervals) < FItSNE.minNumIntervalsMin) || (Integer.parseInt(sMinNumIntervals) > FItSNE.minNumIntervalsMax))
        		sMinNumIntervals = Integer.toString(FItSNE.minNumIntervalsDefault);
        } catch (Exception e) {
        	sMinNumIntervals = Integer.toString(FItSNE.minNumIntervalsDefault);
        }

        String sIntervalsPerInteger = options.get(FItSNE.intervalsPerIntegerOptionName);
        try {
        	if (Double.isNaN(Double.parseDouble(sIntervalsPerInteger)) || (Double.parseDouble(sIntervalsPerInteger) < FItSNE.intervalsPerIntegerMin || Double.parseDouble(sIntervalsPerInteger) > FItSNE.intervalsPerIntegerMax))
        		sIntervalsPerInteger = Double.toString(FItSNE.intervalsPerIntegerDefault);
        } catch (Exception e) {
        	sIntervalsPerInteger = Double.toString(FItSNE.intervalsPerIntegerDefault);
        }

        String sSeed = options.get(FItSNE.seedOptionName);
        try {
        	if ((Integer.parseInt(sSeed) < FItSNE.seedMin) || (Integer.parseInt(sSeed) > FItSNE.seedMax))
        		sSeed = Integer.toString(FItSNE.seedDefault);
        } catch (Exception e) {
        	sSeed = Integer.toString(FItSNE.seedDefault);
        }
        
        BufferedReader rTemplateReader = null;
        try {
			rTemplateReader = new BufferedReader(new InputStreamReader(scriptStream));
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
        
        List<String> opt = new ArrayList<String>();
        opt.add(sAddCellIdToResults);
        opt.add(sDimsOption);
        opt.add(sAlgOption);
        opt.add(sKnnAlgOption);
        opt.add(sPerplexity);
        opt.add(sMaxIter);
        opt.add(sStopLyingIter);
        opt.add(sExaggerationFactor);
        opt.add(sTheta);
        opt.add(sLateEgCoeff);
        opt.add(sStartLateExag);
        opt.add(sNumTrees);
        opt.add(sNumTerms);
        opt.add(sSearchK);
        opt.add(sK);
        opt.add(sSigma);
        opt.add(sMinNumIntervals);
        opt.add(sIntervalsPerInteger);
        opt.add(sSeed);
        opt.add(sNoMomentumDuringExag);
        opt.addAll(parameterNames);
        
        String outFileName = (new StringBuilder()).append(FilenameUtils.fixFileNamePart(sampleName)).append(".FItSNE.").
        		append(FilenameUtils.encodeAlgOptions(opt)).append(".csv").toString();
        File outFile = new File(outputFolder, outFileName);
        outFileName = outFile.getAbsolutePath();

        if(EngineManager.isWindows()) outFileName = outFileName.replaceAll("\\\\", "/");
        if(EngineManager.isWindows()) dataFilePath = dataFilePath.replaceAll("\\\\", "/");
        if(EngineManager.isWindows()) executablePath = executablePath.replaceAll("\\\\", "/");
        
        String scriptLine;
        try {
			while((scriptLine = rTemplateReader.readLine()) != null) 
			{
			    scriptLine = scriptLine.replaceAll("FJ_FITSNE_BINARY_PATH", executablePath);
			    scriptLine = scriptLine.replaceAll("FJ_DATA_FILE_PATH", dataFilePath);
			    scriptLine = scriptLine.replaceAll("FJ_CSV_OUPUT_FILE", outFileName);
			    scriptLine = scriptLine.replaceAll("FJ_PAR_ADD_CELLIDS_TO_RESULT", sAddCellIdToResults);
			    scriptLine = scriptLine.replaceAll("FJ_PAR_PREFIX", prefix);
			    scriptLine = scriptLine.replaceAll("FJ_PAR_DER_PAR_NAME", sDerParName);			    
			    scriptLine = scriptLine.replaceAll("FJ_PAR_PERPLEXITY", sPerplexity);
			    scriptLine = scriptLine.replaceAll("FJ_PAR_MAX_ITER", sMaxIter);
			    scriptLine = scriptLine.replaceAll("FJ_PAR_STOP_LYING_ITER", sStopLyingIter);
			    scriptLine = scriptLine.replaceAll("FJ_PAR_EXAG_FACTOR", sExaggerationFactor);
			    scriptLine = scriptLine.replaceAll("FJ_PAR_THETA", sTheta);
			    scriptLine = scriptLine.replaceAll("FJ_PAR_LATE_EXAG_COEFF", sLateEgCoeff);
			    scriptLine = scriptLine.replaceAll("FJ_PAR_START_LATE_EXAG", sStartLateExag);
			    scriptLine = scriptLine.replaceAll("FJ_PAR_NUM_TREES", sNumTrees);
			    scriptLine = scriptLine.replaceAll("FJ_PAR_NUM_TERMS", sNumTerms);
			    scriptLine = scriptLine.replaceAll("FJ_PAR_SEARCH_K", sSearchK);
			    scriptLine = scriptLine.replaceAll("FJ_PAR_OTHER_K", sK);
			    scriptLine = scriptLine.replaceAll("FJ_PAR_SIGMA", sSigma);
			    scriptLine = scriptLine.replaceAll("FJ_PAR_MIN_NUM_INTERVALS", sMinNumIntervals);
			    scriptLine = scriptLine.replaceAll("FJ_PAR_INTERVALS_PER_INTEGER", sIntervalsPerInteger);
			    scriptLine = scriptLine.replaceAll("FJ_PAR_SEED", sSeed);
			    scriptLine = scriptLine.replaceAll("FJ_PAR_NO_MOM_DURING_EXAG", sNoMomentumDuringExag);
			    

			    scriptLine = scriptLine.replaceAll("FJ_PAR_DIMS", sDimsOption);
			    scriptLine = scriptLine.replaceAll("FJ_PAR_ALG", sAlgOption);
			    scriptLine = scriptLine.replaceAll("FJ_PAR_KNN", sKnnAlgOption);

			    if(scriptLine.contains("FJ_PARAMS_LIST")) {
			    	String parListStr = "";
			    	for (String parName : parameterNames)
			    	{
			    		// We don't want the TIME parameter to be in the parameter list given to the R script
			    		if(parName.compareToIgnoreCase("TIME") != 0) {
			    			if(!parListStr.isEmpty()) parListStr = (new StringBuilder()).append(parListStr).append(",").toString();
			    			parListStr = (new StringBuilder()).append(parListStr).append("\"").append(parName).append("\"").toString();
			    		}
			    	}
			    	scriptLine = scriptLine.replaceAll("FJ_PARAMS_LIST", parListStr);
			    }
			    scriptWriter.append(scriptLine).append('\n');
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
        
        if(rTemplateReader != null) {
        	try { rTemplateReader.close(); }
        	catch (Exception e) { e.printStackTrace(); }
        }
        
        return outFile;
    }

}
